-- if we_byte are not all one, the write cycle takes four cycles to
-- allow read back the old data for those unchanged bytes
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;                                                         
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;
use mylib.DCC_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity vme_if is
    generic (swap_byte : boolean := false; swap_lw : boolean := false);
    Port (
-- VME bus signals
--					 RESETn : in std_logic;
           AM : in  STD_LOGIC_VECTOR (5 downto 0);
					 WRITEn : in std_logic;
           DSn : in  STD_LOGIC_VECTOR (1 downto 0);
					 ASn : in std_logic;
           GAn : in  STD_LOGIC_VECTOR (4 downto 0);
					 GAnP : in std_logic;
					 IACKn : in std_logic;
					 IACKINn : in std_logic;
					 T_DTACK : out std_logic;
					 DTACKn : out std_logic;
					 BERRn : out std_logic;
					 D_DIR : out std_logic;
					 A_DIR : out std_logic;
					 OEn : out std_logic;
					 SYSFAILn : out std_logic;
					 IACKOUTn : out std_logic;
					 IRQ1n : out std_logic;
					 IRQ2n : out std_logic;
					 IRQ4n : out std_logic;
					 IRQ7n : out std_logic;
           A : inout  STD_LOGIC_VECTOR (31 downto 0);
           D : inout  STD_LOGIC_VECTOR (31 downto 0);
-- dip switch
           SW_BAR : in std_logic_vector(4 downto 0);
           SW_use : in std_logic;
-- system clock
           clk : in std_logic;
           inc_addr : in std_logic;
           we_rfifo : in std_logic;
           w_A32_ack : in std_logic;
           rfifo_afull : out std_logic;
           vme_am : out std_logic_vector(5 downto 0);
           vme_read : out std_logic;
           sel_mod : out std_logic;
           r_data : in std_logic_vector(31 downto 0);
           w_data : out std_logic_vector(31 downto 0);
           we_byte : out std_logic_vector(3 downto 0);
--           sys_reset : out std_logic; -- vme system reset
           addr : out std_logic_vector(15 downto 0)
			  );
end vme_if;

architecture my_arch of vme_if is
signal last_word : std_logic := '0';
signal MBLT : std_logic := '0';
signal MBLT_data : std_logic := '0';
--signal vme_reset : std_logic := '0';
--signal vme_reset_sync : std_logic := '0';
signal asn_buf : std_logic := '1';
signal asn_q : std_logic := '1';
signal as : std_logic := '0';
signal viackn : std_logic := '1';
signal a_q : std_logic_vector(31 downto 0) := (others => '0');
signal d_q : std_logic_vector(31 downto 0) := (others => '0');
signal va : std_logic_vector(31 downto 0) := (others => '0');
signal vam : std_logic_vector(5 downto 0) := (others => '0');
signal BAR : std_logic_vector(6 downto 0) := (others => '0');
signal dsn_q : std_logic_vector(1 downto 0) := (others => '0');
signal dsn0_sync : std_logic_vector(1 downto 0) := (others => '0');
signal dsn1_sync : std_logic := '0';
signal sel_mod_i : std_logic := '0';
signal sel_mod_dl : std_logic := '0';
signal w_ack : std_logic := '0';
signal we_byte_i : std_logic_vector(3 downto 0) := (others => '0');
signal en_write : std_logic := '0';
signal rst_rfifo : std_logic := '0';
signal rd_rfifo : std_logic := '0';
signal rfifo_dav : std_logic := '0';
signal rfifo_afull_i : std_logic := '0';
signal rfifo_rp : std_logic_vector(3 downto 0) := (others => '0');
signal r_cycle : std_logic := '0';
signal D32 : std_logic := '0';
signal en_w_ack : std_logic := '0';
signal dtack : std_logic := '0';
signal dtack_dl : std_logic_vector(1 downto 0) := (others => '0');
signal vap : std_logic_vector(17 downto 1) := (others => '0');
signal vme_data : std_logic_vector(63 downto 0) := (others => '0');
signal rfifo_dout : std_logic_vector(31 downto 0) := (others => '0');
signal rfifo_din : std_logic_vector(31 downto 0) := (others => '0');
signal w_data_i : std_logic_vector(31 downto 0) := (others => '0');
attribute IOB: string;
attribute IOB of va,vam,viackn,dsn_q,asn_q,a_q,d_q : signal is "true";
attribute IOBDELAY: string;
attribute IOBDELAY of AM,WRITEn,DSn,ASn,IACKn,A,D : signal is "NONE";
attribute LOC: string;
attribute LOC of i_asn : label is "BUFGMUX_x2y10"; -- global F
attribute INIT: string;
attribute INIT of asn_q : signal is "S";
begin
-- endianess swapping
w_data <= w_data_i when swap_byte = false else w_data_i(7 downto 0) & w_data_i(15 downto 8) & w_data_i(23 downto 16) & w_data_i(31 downto 24);
rfifo_din <= r_data when swap_byte = false else r_data(7 downto 0) & r_data(15 downto 8) & r_data(23 downto 16) & r_data(31 downto 24);
we_byte <= we_byte_i when swap_byte = false else we_byte_i(0) & we_byte_i(1) & we_byte_i(2) & we_byte_i(3);
-- endianess swapping
sel_mod <= sel_mod_i;
vme_am <= vam;
vme_read <= writen;
i_asn: bufg port map(i => asn, o => asn_buf);
rfifo_afull <= rfifo_afull_i;
irq1n <= '1';
irq2n <= '1';
irq4n <= '1';
irq7n <= '1';
iackoutn <= iackinn;
oen <= '0';
sysfailn <= '1';
berrn <= '1';
addr <= vap(17 downto 2);
MBLT <= '1' when vam(1 downto 0) = "00" else '0';
-- synchronize DSn signals
-- synchronize vme_reset
process(en_write,a,d)
begin
  if(en_write = '1')then
    a_q <= a;
		d_q <= d;
  end if;
end process;
process(asn_buf,a,am,iackn)
begin
  if(asn_buf = '1')then
    va <= a;
		vam <= am;
		viackn <= iackn;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
		asn_q <= asn;
		as <= not asn_q;
		if(asn_q = '1')then
			dsn_q <= "11";
		else
			dsn_q <= dsn;
		end if;
    dsn0_sync <= dsn0_sync(0) & dsn_q(0);
    dsn1_sync <= dsn_q(1);
--    vme_reset_sync <= resetn;
--    vme_reset <= not vme_reset_sync;
  end if;
end process;
-- vme system reset signal
--sys_reset <= vme_reset;
-- latching AM and Address lines
BAR(6 downto 2) <= SW_BAR when SW_use = '1' or and_reduce(gan) = '1' or xor_reduce(gan) = ganp else not gan;
BAR(1 downto 0) <= "00";
process(clk)
begin
  if(clk'event and clk = '1')then
    if(as = '0')then
			sel_mod_i <= '0';
		elsif(viackn = '1' and ((vam(5 downto 3) = "001" and BAR = va(31 downto 25)) or (vam(5 downto 3) = "111" and BAR = va(23 downto 17))))then
			sel_mod_i <= '1';
		end if;
		sel_mod_dl <= sel_mod_i;
  end if;
end process;
-- dsn synchronisation and ds rising edge detection
process(clk)
begin
  if(clk'event and clk = '1')then
    if(sel_mod_i = '0')then
			last_word <= not MBLT;
		elsif(en_write = '1' or rd_rfifo = '1')then
			last_word <= not last_word or not MBLT;
		end if;	 
    if(sel_mod_i = '0')then
			MBLT_data <= '0';
		elsif(dtack_dl = "10" and MBLT = '1')then -- the first acknowledge is for address only
			MBLT_data <= '1';
		end if;	 
		if(sel_mod_i = '0')then
			vap <= va(17 downto 1); 
		else
			if(inc_addr = '1' or we_byte_i(3) = '1')then
				vap(10 downto 2) <= vap(10 downto 2) + 1; 
			end if;
			if(va(0) = '1' and dsn0_sync = "01")then
				vap(1) <= not vap(1); 
			end if;
		end if;
    if(dtack = '1')then
			t_dtack <= '0';
		elsif(dtack_dl(0) = '0')then
			t_dtack <= '1';
		end if;
		dtack_dl <= dtack_dl(0) & dtack;
    if(sel_mod_dl = '0' or vam(5) = '1')then
			en_w_ack <= '1';
		elsif(w_A32_ack = '1')then
			en_w_ack <= '1';
		elsif(dtack_dl = "01")then
			en_w_ack <= '0';
		end if;	 
  end if;
end process;
process(clk,dsn_q)
begin
  if(dsn_q = "11")then
    dtackn <= '1';
    dtack <= '0';
  elsif(clk'event and clk = '1')then
    dtackn <= not dtack;
		if(((last_word = '1' and rfifo_dav = '1') or 
			  (((MBLT = '1' and MBLT_data = '0') or((MBLT = '0' or MBLT_data = '1') and writen = '0' and en_w_ack = '1')) and sel_mod_i = '1')) and 
				(dsn0_sync(0) = '0' or dsn1_sync = '0'))then
			dtack <= '1';
		end if;
  end if;
end process; 
-- rescinding DTACKn driven high for two clock cycles(20ns) before going 3-stated;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(sel_mod_i = '0' or rfifo_rp = x"5")then
			rfifo_afull_i <= '0';
		elsif(rfifo_rp = x"6")then
			rfifo_afull_i <= '1';
		end if;
		if(sel_mod_i = '0')then
			D32 <= '0';
		elsif(dsn0_sync(0) = '0' and dsn1_sync = '0' and va(1 downto 0) = "00")then
			D32 <= '1';
		end if;
		if(sel_mod_i = '0' or writen = '1' or w_ack = '1')then
			en_write <= '0';
		elsif((dsn0_sync(0) = '0' or dsn1_sync = '0') and (MBLT = '0' or MBLT_data = '1'))then
			en_write <= '1';
		end if;
		if(dsn0_sync(0) = '1' and dsn1_sync = '1')then
			w_ack <= '0';
		elsif(MBLT = '0' or en_write = '1')then
			w_ack <= '1';
		end if;
-- byte lane select !!! NOTICE byte(0) is the MSB !!!
-- for 16-bit or 8-bit registers, select appropriate we_byte signal
-- according to the address 
		if(en_write = '0')then
			we_byte_i <= x"0";
		elsif(D32 = '1')then
			we_byte_i <= x"f";
		else
			we_byte_i(0) <= not dsn_q(1) and not vap(1);
			we_byte_i(1) <= (not dsn_q(0) and not vap(1)) or not va(0);
			we_byte_i(2) <= (not dsn_q(1) and vap(1)) or not va(0);
			we_byte_i(3) <= not dsn_q(0) and (vap(1) xnor va(0));
		end if;
  end if;
end process;
rst_rfifo <= not sel_mod_i;
i_rfifo: srl_fifo PORT MAP(
		rst => rst_rfifo,
		din => rfifo_din,
		w_strobe => '1',
		we => we_rfifo,
		clk => clk,
		r_strobe => '1',
		re => rd_rfifo,
		dav => rfifo_dav,
		dout => rfifo_dout,
		rp => rfifo_rp
	);
-- D16(D8) data swapping
process(clk)
begin
  if(clk'event and clk = '1')then
    if(rfifo_dav = '1' and rd_rfifo = '0' and ((last_word = '0' and MBLT_data = '1') or 
	    (dsn0_sync = "01" and (D32 = '1' or vap(1) = '1'))))then
			rd_rfifo <= '1';
		else
			rd_rfifo <= '0';
		end if;
		if(swap_lw = true or (last_word = '0' and rd_rfifo = '1'))then
      vme_data(63 downto 32) <= rfifo_dout;
		end if;
		if(swap_lw = true and last_word = '0' and rd_rfifo = '1')then
      vme_data(31 downto 0) <= rfifo_dout;
    elsif(dtack = '0')then
      vme_data(31 downto 16) <= rfifo_dout(31 downto 16);
			if(vap(1) = '0' and va(0) = '1')then
				vme_data(15 downto 0) <= rfifo_dout(31 downto 16);
			else
				vme_data(15 downto 0) <= rfifo_dout(15 downto 0);
			end if;
		end if;
		if(last_word = '0')then
			if(swap_lw = false)then
				w_data_i <= a_q;
			else
				w_data_i <= d_q;
			end if;
		elsif(MBLT = '1' and swap_lw = true)then
			w_data_i <= a_q;
		elsif(va(0) = '0')then
			w_data_i <= d_q;
		else
			w_data_i <= d_q(15 downto 0) & d_q(15 downto 0);
		end if;
		if(dtack = '0')then
			r_cycle <= writen and sel_mod_i and (not MBLT or MBLT_data);
		end if;
  end if;
end process;
d_dir <= r_cycle;
a_dir <= MBLT_data and writen and sel_mod_i; 
a <= (others => 'Z') when MBLT_data  = '0' or writen = '0' or dsn_q(0) = '1' else vme_data(63 downto 32); 
d <= (others => 'Z') when r_cycle = '0' or dsn_q = "11" else vme_data(31 downto 0); 
end my_arch;
